#include "mex.h"
#include "csources/utils.h"
#include "csources/kmeans.h"
#include "csources/DiskUtils.h"

#include <cstring>

using namespace std;

void mexFunction( int nlhs, mxArray *plhs[],
             int nrhs, const mxArray *prhs[] )
{
	if((nlhs == 3) && (nrhs >= 3)) {

		int i=0;

		const mxArray* mx_X	   = prhs[0];
		const mxArray* mx_k        = prhs[1];
		const mxArray* mx_accuracy = prhs[2];

		mxClassID id = mxGetClassID (mx_X);

		int X_columns = mxGetN(mx_X);
		int X_rows    = mxGetM(mx_X);
		double** X = NULL;

		int fileMode = 0;

		if (id == mxCHAR_CLASS) {
			int buflen = X_columns*X_rows*sizeof(mxChar) + 1;
			char* filename = char_array(buflen);

			mxGetString(mx_X, filename, buflen);

			int reorganize = 0;
			int* use_info  = NULL;
			int use_info_size = 0;
			X = load_data (filename, &X_rows, &X_columns, reorganize, use_info, use_info_size, NULL );

			free_array(filename);
			fileMode = 1;
		}
		else {
			double**p_X = (double**) pointer_array(X_columns);

			p_X[0] = mxGetPr (mx_X);
			for(i=0; i < X_columns; i++) p_X[i] = p_X[0] + i*X_rows;

			X = m_create(X_columns, X_rows);
			m_copy(X, p_X, X_columns, X_rows);

			free_array(p_X);
		}

		int codebook_size = (int) *mxGetPr(mx_k);
		double accuracy = *mxGetPr(mx_accuracy); 

		double* labels = NULL;
		double** codebook =  kmeans(X, X_columns, X_rows, codebook_size, accuracy, &labels);

		sadd(labels, labels, 1.0, X_columns);

		mxArray* mx_CODEBOOK = mxCreateNumericMatrix (X_rows, codebook_size, mxDOUBLE_CLASS, mxREAL);
		m_to_vec(mxGetPr(mx_CODEBOOK), codebook, codebook_size, X_rows);

		double rmse = mse (X, X_columns, X_rows, codebook, codebook_size);

		m_free(codebook);

		mxArray* mx_LABELS = mxCreateNumericMatrix(1, X_columns, mxDOUBLE_CLASS, mxREAL);
		memcpy(mxGetPr(mx_LABELS), labels, X_columns*sizeof(double));

		free_array(labels);

		mxArray* mx_mse = mxCreateNumericMatrix(1, 1, mxDOUBLE_CLASS, mxREAL);
		*mxGetPr(mx_mse) = rmse;

		plhs [0] = mx_CODEBOOK;
		plhs [1] = mx_LABELS;
		plhs [2] = mx_mse;

		if (!fileMode) m_free(X);
		else {
			for (i=0; i < X_columns; i++) free_array(X[i]);
			free_array(X);
		}
	}
	else mexPrintf("Wrong number of input/output parameters (%i/%i). \n", nrhs, nlhs);
}

