#include "mex.h"
#include "csources/utils.h"
#include "csources/ngas.h"
#include "csources/kmeans.h"
#include "csources/DiskUtils.h"
#include "csources/Data.h"
#include "csources/MXData.h"

#include <cstring>

using namespace std;

void mexFunction( int nlhs, mxArray *plhs[],
             int nrhs, const mxArray *prhs[] )
{
	if((nlhs >= 1) && (nrhs  >= 3)) {

		int i=0;
		int j=0;

		const mxArray* mx_X	         = prhs[0];
		const mxArray* mx_codebook       = prhs[1];
		const mxArray* mx_num_components = prhs[2];
		const mxArray* mx_noise_level    = (nrhs > 3) ? prhs[3] : NULL;

		int return_residual = (nlhs > 3) ? 1 : 0;
		int return_NR	    = (nlhs > 2) ? 1 : 0;
		int return_N0       = (nlhs > 1) ? 1 : 0;

		mxClassID id = mxGetClassID (mx_X);
	
		int save_memory=0;
		DataDesc* X = MXDataDesc::Create(mx_X, save_memory); 

		int X_columns = X->size();
		int X_rows    = X->ssize();

		int num_codebooks = 1;
		int codebook_size = 0;
			
		int ndims = mxGetNumberOfDimensions(mx_codebook);		

		if (ndims == 2) codebook_size  = mxGetN(mx_codebook);
		else {
			int* dims = int_array(ndims);
			memcpy(dims, mxGetDimensions(mx_codebook), ndims*sizeof(int));
			codebook_size = dims[1];
			num_codebooks = dims[2];
			free_array(dims);
		}

		if ((num_codebooks > 1) && (num_codebooks < X_columns)) {
			PRINTF("Number of codebooks does not match number of samples ... using only the first one ...!!!\n");
			num_codebooks=1;
		}

		double noise_level = (mx_noise_level != NULL) ? *mxGetPr(mx_noise_level) : 0; 
		int num_components = (int) *mxGetPr(mx_num_components);

		double*** codebooks = (double***) pointer_array(num_codebooks);
		double** pcodebook= (double**) pointer_array(codebook_size);
		for (j=0; j < num_codebooks; j++) {

			pcodebook[0] = mxGetPr(mx_codebook) + j*X_rows*codebook_size;
			for(i=0; i < codebook_size; i++) pcodebook[i] = pcodebook[0] + i*X_rows;

			codebooks [j] = m_create(codebook_size, X_rows);
			m_copy(codebooks[j], pcodebook, codebook_size, X_rows);
		}
		free_array(pcodebook);

		double** P = m_create(X_columns, codebook_size);
		double** R = NULL;
		double* N0 = NULL;
		double* NR = NULL;

		double*** r_R = NULL;
		if(return_residual) r_R = &R;

		StandardDataDesc* P_desc = new StandardDataDesc(P, codebook_size, X_columns);  

		omp (P_desc, r_R, &N0, &NR, num_components, noise_level, X, codebooks, num_codebooks, codebook_size);	

		for(i=0; i < num_codebooks;i++) m_free(codebooks[i]);
		free_array(codebooks);

		mxArray* mx_P = mxCreateNumericMatrix(codebook_size, X_columns, mxDOUBLE_CLASS, mxREAL);
		m_to_vec(mxGetPr(mx_P), P, X_columns, codebook_size);

		delete P_desc;
		plhs [0] = mx_P;

		if (return_N0) {
			mxArray* mx_N0 = mxCreateNumericMatrix(1 , X_columns, mxDOUBLE_CLASS, mxREAL);
			memcpy(mxGetPr(mx_N0), N0, X_columns*sizeof(double));
			plhs[1] = mx_N0;
		}
		free_array(N0);
		
		if (return_NR) {
			mxArray* mx_NR = mxCreateNumericMatrix(1 , X_columns, mxDOUBLE_CLASS, mxREAL);
			memcpy(mxGetPr(mx_NR), NR, X_columns*sizeof(double));
			plhs[2] = mx_NR;
		}
		free_array(NR);

		if (return_residual) {
			mxArray* mx_R = mxCreateNumericMatrix(X_rows, X_columns, mxDOUBLE_CLASS, mxREAL);
			m_to_vec(mxGetPr(mx_R), R, X_columns, X_rows);
			m_free(R);
			plhs [3] = mx_R;
		}

		delete X;
	}
	else mexPrintf("Wrong number of input/output parameters (%i/%i). \n", nrhs, nlhs);
}

