% Perform optimized orthogonal matching pursuit in order to find the representation of a number of samples (X) in terms of a 
% given (overcomplete) basis (C)
%
%function [P, N0, NR, R] = OMP(X, C, k, noise_level);
%
%Input:
% X : samples to be encoded (one sample per column)
% C : basis that is used to encode the samples in X (one basis element per column)
% k : maximum number of OMP iterations (i.e., maximum number of non-zero coefficients per sample in P)
%     if noise_level does not equal zero, this is set to dim(X)
%
% noise_level : perform OMP iterations until the norm of the residual drops below noise_level (optional, default : 0)
%
%Output:
% P  : the coefficients (one coefficient vector per column)
% NO : zero-norm of each coefficient vector
% NR : norm of the residual of each sample in X
% R : the residual of each sample
%
%Note: All columns of C have to be set to unit norm
%      X can be obtained from P,R and C by: X = C*P + R
%
function [P, N0, NR, R] = OMP(X, C, k, noise_level);
