#include "mex.h"
#include "csources/utils.h"
#include "csources/ngas.h"
#include "csources/kmeans.h"
#include "csources/DiskUtils.h"
#include "csources/Data.h"
#include "csources/MXData.h"

#include <cstring>

using namespace std;

void mexFunction( int nlhs, mxArray *plhs[],
             int nrhs, const mxArray *prhs[] )
{
	if((nlhs >= 1) && (nrhs >= 4)) {

		int i=0;

		const mxArray* mx_X	         = prhs[0];
		const mxArray* mx_k              = prhs[1];
		const mxArray* mx_epochs         = prhs[2];
		const mxArray* mx_num_components = prhs[3];
		const mxArray* mx_alpha0         = (nrhs > 4) ? prhs[4]   : NULL;
		const mxArray* mx_lambda0        = (nrhs > 5) ? prhs[5]   : NULL;
		const mxArray* mx_alphaf         = (nrhs > 6) ? prhs[6]   : NULL;
		const mxArray* mx_lambdaf        = (nrhs > 7) ? prhs[7]   : NULL;
		const mxArray* mx_noise_level    = (nrhs > 8) ? prhs[8]   : NULL;
		const mxArray* mx_num_track      = (nrhs > 9) ? prhs[9]   : NULL;
		const mxArray* mx_A_orig         = (nrhs > 10) ? prhs[10] : NULL;
		const mxArray* mx_train_index    = (nrhs > 11) ? prhs[11] : NULL;
		const mxArray* mx_t_start        = (nrhs > 12) ? prhs[12] : NULL;
		const mxArray* mx_t_max		 = (nrhs > 13) ? prhs[13] : NULL;
		const mxArray* mx_operation_mode = (nrhs > 14) ? prhs[14] : NULL;
		const mxArray* mx_initial_codebook  = (nrhs > 15) ? prhs[15] : NULL;

		mxClassID id = mxGetClassID (mx_X);

		int save_memory = 0;
		DataDesc* X = MXDataDesc::Create(mx_X, save_memory);
		DataDesc* X_train = NULL;
		DataDesc* X_use = X;

		int* train_index  = NULL;
		int tr_index_size = (mx_train_index) ? mxGetN(mx_train_index)*mxGetM(mx_train_index) : 0;

		if (tr_index_size) { // consider subset
			train_index = int_array(tr_index_size);
			double2int (train_index, mxGetPr(mx_train_index), tr_index_size);

			issub_ip(train_index, 1, tr_index_size); // need zero based indices

			X_train = new SubsetDataDesc(train_index, tr_index_size, X);
			X_use = X_train;
		}

		int X_columns = X_use->size();
		int X_rows    = X_use->ssize();

		int codebook_size  = (int) *mxGetPr(mx_k);
		int epochs         = (int) *mxGetPr(mx_epochs); 
		int num_components = (int) *mxGetPr(mx_num_components); 

		double alpha0  = (mx_alpha0 != NULL)  ? ((!mxIsEmpty(mx_alpha0)) ?  *mxGetPr(mx_alpha0) : -1)  : -1;
		double lambda0 = (mx_lambda0 != NULL) ? ((!mxIsEmpty(mx_lambda0)) ? *mxGetPr(mx_lambda0) : -1) : -1;

		double alphaf  = (mx_alphaf  != NULL) ? ((!mxIsEmpty(mx_alphaf)) ? *mxGetPr(mx_alphaf) : -1) : -1;
		double lambdaf = (mx_lambdaf != NULL) ? ((!mxIsEmpty(mx_lambdaf)) ? *mxGetPr(mx_lambdaf) : -1) : -1;

		double noise_level = (mx_noise_level != NULL) ? ((!mxIsEmpty(mx_noise_level)) ? *mxGetPr(mx_noise_level) : 0.0) : 0.0;

		int num_track = 0;
		if (mx_num_track) if (!mxIsEmpty(mx_num_track)) num_track = (int) *mxGetPr(mx_num_track);

		PRINTF("num_track: %i\n", num_track);

		double** A_orig = NULL;
		if(mx_A_orig) {
			if (!mxIsEmpty(mx_A_orig)) {
				double** pA_orig = (double**) pointer_array(X_columns);
				pA_orig[0] = mxGetPr(mx_A_orig);
				for(i=1; i < X_columns; i++) pA_orig[i] = pA_orig[0] + i*codebook_size;	

				A_orig = m_create(X_columns, codebook_size);
				m_copy(A_orig, pA_orig, X_columns, codebook_size);	

				free_array(pA_orig);
			}
		}

		int operation_mode = (mx_operation_mode != NULL) ? ((!mxIsEmpty(mx_operation_mode)) ? (int) *mxGetPr(mx_operation_mode) : OPERATION_MODE_BATCH) : OPERATION_MODE_BATCH;

		PRINTF("operation mode: %i\n", operation_mode);

		double* E    = NULL;
		double* N0   = NULL;
		double* AE   = NULL;
		if ((num_track > 0) && (A_orig != NULL) && (operation_mode == OPERATION_MODE_BATCH)) {
			E  = double_array(num_track);
			N0 = double_array(num_track);
			AE = double_array(num_track);
		}

		double*** track_codebook = NULL;
		if ((operation_mode != OPERATION_MODE_BATCH) && (num_track > 0)) track_codebook = (double***) pointer_array(num_track);

		double** initial_codebook = NULL;

		if(mx_initial_codebook) {
			if ((mxGetM(mx_initial_codebook) == X_rows) && (mxGetN(mx_initial_codebook) == codebook_size)) {
				double** pinitial_codebook= (double**) pointer_array(codebook_size);
				pinitial_codebook[0] = mxGetPr(mx_initial_codebook);
				for(i=0; i < codebook_size; i++) pinitial_codebook[i] = pinitial_codebook[0] + i*X_rows;

				initial_codebook = m_create(codebook_size, X_rows);
				m_copy(initial_codebook, pinitial_codebook, codebook_size, X_rows);

				free_array(pinitial_codebook);
			}
			else PRINTF("Initial codebook has wrong dimensions !!!!\n");
		}

		int t_start = (mx_t_start != NULL) ? (int) *mxGetPr(mx_t_start) : 0;
		int t_max   = (mx_t_max   != NULL) ? (int) *mxGetPr(mx_t_max)   : 0;

		PRINTF("t_start: %i, t_max: %i\n", t_start, t_max);

		double** codebook = spngas(X_use, codebook_size, epochs, num_components,
					alpha0, lambda0, alphaf, lambdaf, noise_level, t_start, t_max,
					E,N0, AE, A_orig, track_codebook,
					num_track, operation_mode, initial_codebook);

		mxArray* mx_CODEBOOK = mxCreateNumericMatrix (X_rows, codebook_size, mxDOUBLE_CLASS, mxREAL);
		m_to_vec(mxGetPr(mx_CODEBOOK), codebook, codebook_size, X_rows);
		m_free(codebook);

		plhs [0] = mx_CODEBOOK;

		if (operation_mode == OPERATION_MODE_BATCH) {
			if ((nlhs > 1) && (E != NULL)) {
				mxArray* mx_E = mxCreateNumericMatrix (1, num_track, mxDOUBLE_CLASS, mxREAL);
				memcpy(mxGetPr(mx_E), E, num_track*sizeof(double));
				plhs[1] = mx_E;
			}

			if ((nlhs > 2) && (N0 != NULL)) {
				mxArray* mx_N0 = mxCreateNumericMatrix (1, num_track, mxDOUBLE_CLASS, mxREAL);
				memcpy(mxGetPr(mx_N0), N0, num_track*sizeof(double));
				plhs[2] = mx_N0;
			}

			if ((nlhs > 3) && (AE != NULL)) {
				mxArray* mx_AE = mxCreateNumericMatrix (1, num_track, mxDOUBLE_CLASS, mxREAL);
				memcpy(mxGetPr(mx_AE), AE, num_track*sizeof(double));
				plhs[3] = mx_AE;
			}
		}
		else {
			if ((nlhs > 1) && (num_track) && (track_codebook != NULL)) {
				int ndim = 3;
				int dims [3];
				dims[0] = X_rows;
				dims[1] = codebook_size;
				dims[2] = num_track;
		
				mxArray* mx_CODEBOOKS = mxCreateNumericArray(ndim, dims, mxDOUBLE_CLASS, mxREAL);

				double* pret = mxGetPr(mx_CODEBOOKS);
				for (i=0;i < num_track;i++) { // copy the codebook track to the MATLAB structure
					m_to_vec(pret, track_codebook[i], codebook_size, X_rows);
					pret += X_rows*codebook_size;
				}

				plhs[1] = mx_CODEBOOKS;

				for (i=0;i < num_track;i++) m_free(track_codebook[i]);
				free_array(track_codebook);
			}
		}

		if(A_orig) m_free(A_orig);

		if (E  != NULL) free_array(E);
		if (N0 != NULL) free_array(N0);
		if (AE != NULL) free_array(AE);

		if (train_index) free_array(train_index);
	
		delete X;
		if (X_train) delete X_train;

		if (initial_codebook) m_free(initial_codebook);
	}
	else mexPrintf("Wrong number of input/output parameters (%i/%i). \n", nrhs, nlhs);
}

