% The Sparse Coding Neural Gas (SCNG) algorithm [1],[2],[3].
%
% Standard usage:
%  C = SCNG(X, codebook_size, epochs, num_components);
%
%Input:
% X                : data samples, one sample per column (or name of file containing one sample per row in ascii format)
% codebook_size    : number of basis functions to be used to represent the data samples 
%                    via orthognal matching pursuit (OMP)
% epochs           : number of learning epochs (train_steps = epochs * number of data samples)
% num_components   : number of basis functions that are allowed to represent a single sample
% alpha0           : initial learning rate (optional, default 0.5)
% lambda0          : initial neighbourhood size (optional, default 0.5 * codebook_size)
% alphaf           : final learning rate (optional, default 0.005)
% lambdaf          : final neighbourhood size (optional, 0.01)
% noise_level	   : stop OMP if the norm of the current residual drops below noise_level 
%                    (to be used for BSS [2], optional, default: 0)
% num_track        : BATCH MODE -> defines how often the reconstruction error and the zero
%                    norm of the coefficients should be evaluated during
%                    learning,
%                    ONLINE MODE -> defines how often the current codebook should be saved
%                    (optional, default: 0 -> off)
% A_orig           : original coefficients of the linear combinations 
%                    (required if num_track > 0 & BATCH MODE !!!, optional, default [])
% train_index      : use only a subset of all samples to learn the basis
%                    (optional, default [])
% t_start 	       : start learning at time step t_start 
%                    (required for online learning, optional, default 0)
% t_max		       : maximum number of training steps 
%                    (required for online learning, optional, default t_max=epochs*number of data samples)
% operation_mode   : 0 -> BATCH, 1 -> ONLINE [3] (optional, default BATCH)
% initial_codebook : use inital_codebook as starting point for learning 
%                    (to be used for online learning, optional, default [])
%
%Output:
%  C : codebook that was learned 
%  E : reconstruction error evaluated num_track times during learning (BATCH MODE) 
%	or
%track_codebook : num_track snapsshots of the codebook taken during learning
%
%  N0 : mean zero norm of the coefficients evaluated E_length times during learning (BATCH MODE only)
%  AE : representation error evaluated E_length times during learningi (BATCH MODE only)
%
% Complete signature:
%function [C,E, N0, AE]=SCNG(X, codebook_size, epochs, num_components, alpha0, lambda0, alphaf, lambdaf, noise_level, 
%	                     num_track, A_orig, train_index, t_start, t_max, operation_mode, initial_codebook);
%
%-----------
% [1] Kai Labusch, Erhardt Barth, and Thomas Martinetz. Sparse Coding Neural Gas: Learning of Overcomplete Data Representations. 
%     Neurocomputing, 72(7-9):1547-1555, 2009.
% [2] Kai Labusch, Erhardt Barth, and Thomas Martinetz. Demixing Jazz-Music: Sparse Coding Neural Gas for the Separation of 
%     Noisy Overcomplete Sources. Neural Network World, 19(5):561-579, 2009.
% [3] Kai Labusch, Erhardt Barth, and Thomas Martinetz. Approaching the Time Dependent Cocktail Party Problem with Online Sparse 
%     Coding Neural Gas. In J.C. Principe and R. Miikkulainen, editors, Advances in Self-Organizing Maps - WSOM 2009,
%     7th International Workshop, St. Augustine, Fl, USA, June 2009, volume 5629 of Lecture Notes in Computer Science, 
%     pages 145-153. Springer, 2009.
%
function [C,E, N0, AE]=SCNG(X, codebook_size, epochs, num_components, alpha0, lambda0, alphaf, lambdaf, noise_level,...
			    E_length, A_orig, train_index, t_start, t_max, operation_mode, initial_codebook)

