#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <regex.h>
#include <cmath>
#include <errno.h>

#include "utils.h"
#include "Data.h"
#include "DiskUtils.h"

void DataDesc::GetAsMatrix(double*** ret) {
	double** ptemp = (double**) pointer_array(size());

	int* itemp = int_array(size());
	for (int i=0; i < size();i++) itemp[i]=i;

	GetSamples (ptemp, itemp, size());

	*ret = m_create(size(), ssize());

	m_copy(*ret, ptemp, size(), ssize());

	free_array(ptemp);
	free_array(itemp);
}


///////////////////////7
/////
////////////////////////

MeanRemover::MeanRemover(double* m, DataDesc* D, int save_memory) : ProxyDataDesc(D), mean(NULL), data(D), save_memory(save_memory), removed(NULL), I_removed(NULL), num_removed(0) {
	mean = double_array(D->ssize());	
	memcpy(mean, m, D->ssize()*sizeof(double));

	removed = int_array(D->size());
	memset(removed, 0, D->size()*sizeof(int));

	I_removed = int_array(D->size());
	memset(I_removed, 0, D->size()*sizeof(int));
}
	
MeanRemover::~MeanRemover(void) {
	free_array(mean);
	free_array(removed);
	free_array(I_removed);
}

void MeanRemover::GetSample  (double** p, int i) { 
	ProxyDataDesc::GetSample  (p, i);

	if (!removed[i]) {
		sub(*p, *p, mean, data->ssize());
		removed[i]=1;
		I_removed[num_removed] = i;
		num_removed++;
	} 
}

void MeanRemover::GetSamples (double** P, int* I, int len) { 
	ProxyDataDesc::GetSamples (P, I, len);

	for (int i=0; i < len; i++) {
		if (!removed[I[i]]) {
			sub(P[i], P[i], mean, data->ssize());
		
			removed[I[i]]=1;
			I_removed[num_removed]=I[i];
			num_removed++;
		}
	} 
}

int MeanRemover::CopySample (double* x, int** NZ, int i) { 

	int ret = ProxyDataDesc::CopySample (x, NZ, i);

	if (NZ != NULL) {
		if (*NZ != NULL) {

			double* tmp = double_array(ret);
		
			get_i(tmp, mean, *NZ, ret);
			sub(x,x,tmp, ret);

			free_array(tmp);
		}
		else sub(x,x,mean, data->ssize()); 
	}

	return ret;
}

void MeanRemover::Flush(void) {
	if (save_memory) {

		for (int i=0; i < num_removed; i++) removed[I_removed[i]] = 0;
		num_removed=0;

		//memset(removed, 0, data->size()*sizeof(int));
		ProxyDataDesc::Flush();
	}
}

////////////////////////////////////////////////////////


StandardDataDesc::StandardDataDesc (double** X, int esize, int nentrys, int isOwner) : D(X), entry_size(esize), size(nentrys), isOwner(isOwner) {
}

StandardDataDesc::~StandardDataDesc(void) {
	if(isOwner)m_free(D);
}

int StandardDataDesc::ssize(void) {
	return entry_size;
}

int StandardDataDesc::num_samples (void) {
	return size;
}

int StandardDataDesc::Count(void) {
	return size*entry_size;
}

void StandardDataDesc::GetSample(double** x, int i) {
	*x = D[i];
}

void StandardDataDesc::GetSamples (double** P, int* I, int len) {
	pget_i(P, D, I, len);
}

int StandardDataDesc::CopySample (double* x, int** NZ, int i) {
	if (NZ != NULL) *NZ = NULL;

	memcpy(x, D[i], sizeof(double)*entry_size);

	return entry_size;
}

void StandardDataDesc::PutSample(double* x, int* NZ, int NZ_len, int i) {
	if (NZ !=  NULL) {
		double* tmp = double_array(NZ_len);
		memset(D[i], 0, entry_size*sizeof(double));
		get_i(tmp, x, NZ, NZ_len);
		set_i(D[i], tmp, NZ, NZ_len); 
		free_array(tmp);
	}
	else memcpy(D[i], x, entry_size*sizeof(double));
}

void StandardDataDesc::Flush(void) {
}

void StandardDataDesc::getAbsIndex (int* index) { 
	for (int i=0; i < size; i++) index[i] = i; 
}

void StandardDataDesc::getAbsIndex (int* index, int* I, int len) { 

	memcpy(index, I, len*sizeof(int)); 
};

///////////////////////////////////////////////
///
///////////////////////////////////////////////

ImageDataDesc::ImageDataDesc(double** image, int nentrys, int esize, int tile_nentrys, int tile_esize, int random_tile_count, int isOwner) {
}

ImageDataDesc::~ImageDataDesc(void) {
}

int ImageDataDesc::ssize(void) {
	return -1;
}

int ImageDataDesc::num_samples (void) {
	return -1;
}

int ImageDataDesc::Count(void) {
	return -1;
}

void ImageDataDesc::GetSample(double** x, int i) {
}

void ImageDataDesc::GetSamples (double** P, int* I, int len) {
}

int ImageDataDesc::CopySample (double* x, int** NZ, int i) {
	return -1;
}

void ImageDataDesc::PutSample(double* x, int* NZ, int NZ_len, int i) {
}

void ImageDataDesc::Flush(void) {
}

void ImageDataDesc::getAbsIndex (int* index) {
}

void ImageDataDesc::getAbsIndex (int* index, int* I, int len) {
}

///////////////////////////////////////////////////////////7
//////////////////////////////////////////////////////////7
//
/////////////////////////////////////////////////////////////7
//////////////////////////////////////////////////////////7
//
//

MemStore::MemStore(int size, int esize) : size(size), esize(esize){
	in_mem = (double**) pointer_array(size);	
	memset(in_mem, 0, size*sizeof(double**));

	i_in_mem = int_array(size);
	memset(i_in_mem, 0, size*sizeof(int));
		
	in_mem_size=0;
}

MemStore::~MemStore (void) {
	for(int i=0; i < in_mem_size; i++) free_array(in_mem[i_in_mem[i]]);	
	free_array(in_mem);
	free_array(i_in_mem);
}

double* MemStore::ref(int i) {
	// mutex sperren

	if (!is_in_mem(i)) {
		in_mem[i] = double_array(esize);
		memset(in_mem[i], 0, esize*sizeof(double));
		i_in_mem[in_mem_size] = i;
		in_mem_size++;	
	}

	// mutex entsperren

	return in_mem[i];
}
int MemStore::is_in_mem(int i) {
	return (int) (in_mem[i] != NULL);
}

void MemStore::flush(void) {
	for(int i=0; i < in_mem_size; i++) { 
		free_array(in_mem[i_in_mem[i]]);	
		in_mem[i_in_mem[i]] = NULL;
	}

	in_mem_size = 0;
}

void MemStore::set_entry_size(int new_size) {
	esize = new_size;
}

////////////////////////////////////////////////////////////////////////////
//'////////////////////////////////////////////////////////////////////////


SparseDataDesc::SparseDataDesc (char* directory) {

	char buf[1024];
	sprintf(buf,"%s/%s", directory, "size.dat");

	int esize     =0;
	int num_entrys=0;
	int reorganize=1;

	int*use_info      =NULL; // unused
	int use_info_size =0;    // unused
	double* file_index=NULL; // unused

	// load size of matrix from size.dat
	double** size_tmp  = load_data (buf, &esize, &num_entrys, reorganize, use_info, use_info_size, file_index);
	double*  size_tmp2 = double_array(esize*num_entrys);

	// convert size matrix to vector
	m_to_vec (size_tmp2, size_tmp, num_entrys, esize);

	nsamples   = (int) size_tmp2[0]; // first entry is number of samples
	entry_size = (int) size_tmp2[1]; // second entry is size of each sample

	PRINTF("SparseDataDesc: Loading sparse matrix of size (%i x %i)...\n", nsamples, entry_size);

	m_free(size_tmp); // size matrix is not used anymore
	free_array(size_tmp2);

	D              = (double**) pointer_array(nsamples);
	active_dims    = (int**)    pointer_array(nsamples);
	active_dim_len = int_array(nsamples);

	memset(D             , 0, nsamples*sizeof(double**));
	memset(active_dims   , 0, nsamples*sizeof(int**));
	memset(active_dim_len, 0, nsamples*sizeof(int));

	store = new MemStore(nsamples, entry_size);

	for(int i=0; i < nsamples; i++) {

		sprintf(buf,"%s/%i.dat", directory, i+1);

		double** tmp = load_data (buf, &esize, &num_entrys, reorganize, use_info, use_info_size, file_index);
		
		if (tmp != NULL) {
			D[i]           = double_array(esize);
			active_dims[i] = int_array(esize);
	
			memcpy(D[i], tmp[0], esize*sizeof(double));
			double2int(active_dims[i], tmp[1], esize);

			issub(active_dims[i], active_dims[i], 1, esize);

			active_dim_len[i] = esize;

			m_free(tmp);
		}
		else {
			D[i] = NULL;
			active_dims[i]    = NULL;
			active_dim_len[i] = 0;
		}
	}
}

SparseDataDesc::SparseDataDesc (int nentrys, int esize) : nsamples(nentrys), entry_size(esize) {
	D              = (double**) pointer_array(nsamples);
	active_dims    = (int**)    pointer_array(nsamples);
	active_dim_len = int_array(nsamples);

	store = new MemStore(nsamples, entry_size);

	memset(D          , 0, nsamples*sizeof(double**));
	memset(active_dims, 0, nsamples*sizeof(int**));
	memset(active_dim_len, 0, nsamples*sizeof(int));
}

int SparseDataDesc::Count(void) {
	return isum(active_dim_len, nsamples); 
}

void SparseDataDesc::Flush(void) {
	store->flush();
}

void SparseDataDesc::save(char* directory) {

	mode_t mode = S_IRUSR | S_IWUSR | S_IXUSR;

        int ret =  mkdir (directory, mode);

	if ((ret == 0) || (ret == EEXIST)) {
		if (ret == EEXIST) PRINTF("directory %s already exists ... overwriting ...\n", directory);
		else PRINTF("writing to directory %s ...\n", directory);

		char size_file[1024];
		sprintf(size_file, "%s/size.dat", directory);

		FILE* size_fid = fopen(size_file, "w+");
		fprintf(size_fid, "%i %i", nsamples, entry_size);
		fclose(size_fid);

		double** M = m_create(2, entry_size); 

		int i=0;
		for (i=0;i < nsamples; i++) {
			
			if((D[i] != NULL) && (active_dim_len[i] > 0)){
				char filename [1024];
				sprintf(filename, "%s/%i.dat", directory, i+1);
//				PRINTF("Writing file %s ...\n", filename);

				memcpy(M[0],D[i], active_dim_len[i]*sizeof(double));
				int2double(M[1], active_dims[i], active_dim_len[i]);
				sadd(M[1], M[1], 1.0, active_dim_len[i]);

				if (!save_data (filename, M, active_dim_len[i], 2)) PRINTF("SparseDataDesc::save: Could not write file %s ....! \n", filename);
			}
		}

		m_free(M);
	}
	else PRINTF("SparseDataDesc::save : Could not create directory %s !!!\n", directory);
}

void SparseDataDesc::get_matrix(double*** M, int* nentrys, int* esize) {
	*nentrys = nsamples;
	*esize   = entry_size;

	*M = m_create(nsamples, entry_size);
	m_zero(*M, nsamples, entry_size);

	int i=0;
	for(i=0; i < nsamples; i++) if (D[i] != NULL) set_i((*M)[i], D[i], active_dims[i], active_dim_len[i]);
}

void SparseDataDesc::mean (double* data_mean) {
	
	double* data_tmp = double_array(entry_size);

	memset(data_mean, 0, entry_size*sizeof(double));

	double* ne_count = double_array(entry_size);
	memset(ne_count, 0, entry_size*sizeof(double));

	int* I = int_array(entry_size);

	int i=0;
	for(i=0; i < nsamples; i++) {
		if(D[i] != NULL) {
			set_i(data_tmp, D[i], active_dims[i], active_dim_len[i]);

			add_ip_i (data_mean, data_tmp, active_dims[i], active_dim_len[i]);
			sadd_ip_i(ne_count, 1.0, active_dims[i], active_dim_len[i]); // count number of non-zeros at each position
		}
	}

	int zero_dims = find_eq (I, ne_count, 0, entry_size); // prevent division by zero
	assign_i(ne_count, I, 1.0, zero_dims);

	divide(data_mean, data_mean, ne_count, entry_size); // divide each dimension by its number of non-zero entries

	free_array(ne_count);
	free_array(I);
	free_array(data_tmp);
}

void SparseDataDesc::sample_mean(double* smean) {

	memset(smean, 0, sizeof(double)*nsamples);

	int i=0;
	for(i=0;i < nsamples;i++) {
		if (D[i] != NULL) {
			smean[i] = ::mean(D[i], active_dim_len[i]);
		}
	}
}

void SparseDataDesc::ssub(double* sum) {
	int i=0;
	
	for(i=0; i < nsamples; i++) {
		if (D[i] != NULL) {
			::ssub(D[i], D[i], sum[i], active_dim_len[i]);
		}
	}
}

void SparseDataDesc::sub(double* sum) {
	int i=0;
	
	double* data_tmp = double_array(entry_size);	

	for(i=0; i < nsamples; i++) {
		if (D[i] != NULL) {
			get_i(data_tmp, sum, active_dims[i], active_dim_len[i]);
			::sub(D[i], D[i], data_tmp, active_dim_len[i]); // remove_mean
		}
	}

	free_array(data_tmp);
}

void SparseDataDesc::smultiply(double s) {
	int i=0;

	for(i=0;i < nsamples; i++) {
		if(D[i] != NULL) {
			::smultiply(D[i], D[i], s, active_dim_len[i]);
		}
	}	
}

double SparseDataDesc::mean_var(void) {
	int i=0;

	double* tmp = double_array(entry_size);

	double N = (double) isum(active_dim_len, nsamples);

	double mv = 0.0;
	
	for(i=0; i < nsamples;i++) {
		if(D[i] != NULL) {
			::multiply(tmp,D[i],D[i], active_dim_len[i]);

			mv += sum(tmp, active_dim_len[i]);
		}
	}

	free_array(tmp);

	return mv/(N-1);
}

void SparseDataDesc::get_norm(double* norm) {

	memset(norm, 0, nsamples*sizeof(double));

	int i=0;
	for (i=0;i < nsamples;i++) {
		if(D[i] != NULL) {
			norm[i] = sqrt(scp(D[i],D[i], active_dim_len[i]));
			if(norm[i]==0) norm[i]=1;
		}
		else norm[i] = 1.0;
	}
}

void SparseDataDesc::smultiply(double* fac) {
	int i=0;
	for (i=0;i < nsamples;i++) {
		if (D[i] != NULL) ::smultiply(D[i], D[i], fac[i], active_dim_len[i]);
	}
}

SparseDataDesc::SparseDataDesc (int num_samples, int esize, double** D_tmp, int** active_dims_tmp, int* active_dim_len_tmp) : nsamples(num_samples), entry_size(esize),
												D(D_tmp), active_dims(active_dims_tmp), active_dim_len(active_dim_len_tmp) {

	store = new MemStore(num_samples, esize);
}

SparseDataDesc::~SparseDataDesc (void) {
	int i=0;

	for(i=0;i < nsamples; i++) {
		if (D[i] != NULL) free_array(D[i]);
		if (active_dims[i] != NULL) free_array(active_dims[i]);
	}

	free_array(D);
	free_array(active_dims);
	free_array(active_dim_len);

	delete store;
}

SparseDataDesc* SparseDataDesc::transpose(void) {

	int i=0;
	int j=0;

	int* ne_count = int_array(entry_size);
	memset(ne_count, 0, entry_size*sizeof(int));

	for(i=0;i < nsamples; i++) {
		if (active_dims[i] != NULL) isadd_ip_i(ne_count, 1, active_dims[i], active_dim_len[i]); /* count number of non-zeros at each position */
	}

	int* used_counter = int_array(entry_size);
	memset(used_counter, 0, entry_size*sizeof(int));

	double** D_tp        = (double**) pointer_array(entry_size);
	int** active_dims_tp = (int**)    pointer_array(entry_size);

	memset(D_tp, 0, entry_size*sizeof(double**));
	memset(active_dims_tp, 0, entry_size*sizeof(int**));

	for(i=0; i < entry_size;i++) { /* allocate memory for transposed version */
		D_tp[i]            = double_array(ne_count[i]);
		active_dims_tp [i] = int_array(ne_count[i]);
	}

	for (i=0; i < nsamples;i++) { /* actually transpose the matrix */

		if (D[i] != NULL) {
			int*    cur_dims   = active_dims[i];
			double* cur_values = D[i];

			for(j=0; j < active_dim_len[i]; j++) {

				int pos = cur_dims[j];

				D_tp          [pos][used_counter[pos]] = cur_values[j];
				active_dims_tp[pos][used_counter[pos]] = i;
		
				used_counter[pos]++;	
			}
		}
	}

	for (i=0; i < entry_size; i++) {
		if (D_tp[i] != NULL) isort(D_tp[i], active_dims_tp[i], ne_count[i]);
	}	

	free_array(used_counter);

	return new SparseDataDesc(entry_size, nsamples, D_tp, active_dims_tp, ne_count);
}

SparseDataDesc* SparseDataDesc::copy(void) {

	double** D_cp          = (double**) pointer_array(nsamples);
	int** active_dims_cp   = (int**)    pointer_array(nsamples);
	int* active_dim_len_cp = int_array(nsamples);	

	memset(D_cp             , 0, nsamples*sizeof(double**));
	memset(active_dims_cp   , 0, nsamples*sizeof(int**));
	memset(active_dim_len_cp, 0, nsamples*sizeof(int));

	memcpy(active_dim_len_cp, active_dim_len, nsamples*sizeof(int));

	int i=0;
	for (i=0; i < nsamples;i++) {
		if(D[i] != NULL) {
			D_cp[i] = double_array(active_dim_len[i]);
			memcpy(D_cp[i], D[i], active_dim_len[i]*sizeof(double));

			active_dims_cp[i] = int_array(active_dim_len[i]);
			memcpy(active_dims_cp[i], active_dims[i], active_dim_len[i]*sizeof(int));
		}
	}

	return new SparseDataDesc(nsamples, entry_size, D_cp, active_dims_cp, active_dim_len_cp);
}

void SparseDataDesc::put(SparseDataDesc* src) {
	
	if ((src->nsamples != nsamples) || (src->entry_size != entry_size)) {
		PRINTF(" SparseDataDesc::put: matrix dimensions do not match !!!\n");
		return;
	}
	
	double* data_tmp = double_array(entry_size);

	int i=0;
	for (i=0;i < nsamples; i++) {
		//memset(data_tmp, 0, entry_size*sizeof(double));


		if (D[i] != NULL) set_i(data_tmp, D[i], active_dims[i], active_dim_len[i]);

		if (src->D[i] != NULL) set_i(data_tmp, src->D[i], src->active_dims[i], src-> active_dim_len[i]);

		int* dim_union = int_array(active_dim_len[i] + src->active_dim_len[i]);
		int* un        = int_array(active_dim_len[i] + src->active_dim_len[i]); 

		memcpy(dim_union                    , active_dims[i]     , active_dim_len[i]*sizeof(int)); 

		memcpy(dim_union + active_dim_len[i], src->active_dims[i], src->active_dim_len[i]*sizeof(int));

		int new_active_len = iunique(un, dim_union, active_dim_len[i] + src->active_dim_len[i]);

		if (D[i] != NULL) {
			free_array(D[i]);
			free_array(active_dims[i]);
		}

		D[i]           = double_array(new_active_len);
		active_dims[i] = int_array(new_active_len);

		memcpy(active_dims[i], un, new_active_len*sizeof(int)); 
		if (active_dims[i] != NULL) get_i(D[i], data_tmp, active_dims[i], new_active_len);
		active_dim_len[i] = new_active_len;

		free_array(dim_union);
		free_array(un);
	}

	free_array(data_tmp);
}

double SparseDataDesc::count_entrys(double value) {

	double ret = 0;

	int* I = int_array(entry_size);

	int i=0;
	for (i=0; i < nsamples; i++) {
		if (D[i] != NULL) ret += (double) find_eq(I, D[i], value, active_dim_len[i]);	
	}

	free_array(I);
		
	return ret;
}

void SparseDataDesc::delete_entrys(double value) {

	int* I      = int_array(entry_size);

	int i=0;
	for (i=0; i < nsamples; i++) {

		if (D[i] != NULL) {
			int new_len = find_neq(I, D[i], value, active_dim_len[i]);	

			if (new_len != active_dim_len[i]) {
				double* d = double_array(new_len);
				int*   ad = int_array(new_len);

				get_i(d, D[i], I, new_len);
				iget_i(ad, active_dims[i], I, new_len); 

				free_array(D[i]);
				D[i] = d;
				free_array(active_dims[i]);
				active_dims[i] = ad;

				active_dim_len[i] = new_len;
			}
		}
	}

	free_array(I);
}

void SparseDataDesc::get(SparseDataDesc* dest) {
	if ((dest->nsamples != nsamples) || (dest->entry_size != entry_size)) {
		PRINTF(" SparseDataDesc::get: matrix dimensions do not match !!!\n");
		return;
	}

	double* data_tmp = double_array(entry_size);

	int i=0;
	for (i=0;i < nsamples; i++) {
		if (D[i] != NULL) {
			memset(data_tmp, 0, entry_size*sizeof(double));

			set_i(data_tmp  , D[i], active_dims[i], active_dim_len[i]);
			if (dest->D[i] != NULL) get_i(dest->D[i], data_tmp, dest->active_dims[i], dest->active_dim_len[i]);
		}
	}

	free_array(data_tmp);
}

int SparseDataDesc::ssize(void) {
	return entry_size;
}

int SparseDataDesc::num_samples (void) {
	return nsamples;
}

void SparseDataDesc::GetSample(double** x, int i) {

	if (store->is_in_mem(i)) *x = store->ref(i);
	else {
		double* data_tmp = store->ref(i);

		memset(data_tmp, 0, entry_size*sizeof(double));
		if (D[i] != NULL) set_i(data_tmp, D[i], active_dims[i], active_dim_len[i]);

		*x = data_tmp;
	}
}

void SparseDataDesc::GetSamples (double** P, int* I, int len) {
	int i=0;
	for (i=0; i < len; i++)  GetSample(P + i, I[i]);
}

int SparseDataDesc::CopySample (double* x, int** NZ, int i) {
	memcpy(x, D[i], active_dim_len[i]*sizeof(double));	
	if(NZ != NULL) *NZ = active_dims[i];

	return active_dim_len[i];
}

void SparseDataDesc::PutSample(double* x, int* NZ, int NZ_len, int i) {

	if(NZ_len != active_dim_len[i]){
		if (D[i] != NULL) free_array(D[i]);
		D[i] = double_array(NZ_len);

		if (active_dims[i] != NULL) free_array(active_dims[i]);
		active_dims[i] = int_array(NZ_len);

		active_dim_len[i] = NZ_len;
	}

	memcpy(D[i]          , x , NZ_len*sizeof(double));
	memcpy(active_dims[i], NZ, NZ_len*sizeof(int));
}

void SparseDataDesc::getAbsIndex (int* index) { 
	for (int i=0; i < size(); i++) index[i] = i; 
}

void SparseDataDesc::getAbsIndex (int* index, int* I, int len) { 
	memcpy(index, I, len*sizeof(int)); 
}


/////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////

DiskDataDesc::DiskDataDesc(char* fname, int save_mem) : save_memory(save_mem) {
	filename = char_array(strlen(fname)+1);
	sprintf(filename,"%s",fname);

	// create file index in order to provide faster access to the data
	file_index = create_file_index (filename, &file_index_len);

	int tmp_index  = 1; // just load the first line of the file to determine entry size 
	int tmp        = 0;
	int output     = 0;
	double** tmp_m = load_data (filename, &entry_size, &tmp, 1, &tmp_index, 1 , file_index, output);
	m_free(tmp_m);

	store = new MemStore(file_index_len, entry_size);
}

DiskDataDesc::~DiskDataDesc (void) {

	free_array(filename);

	free_array(file_index);

	delete store;
}

void DiskDataDesc::GetSample  (double** p, int i) {
	if (store->is_in_mem(i)) *p = store->ref(i);
	else {
		*p = store->ref(i);


		int reorder   = 1;
		int tmp=0;
		int uinfo     = i;
		int uinfo_len = 1;
		int esize     = entry_size;
		int output    = 0;

		double** tmp_m = load_data (filename, &entry_size, &tmp, reorder, &uinfo, uinfo_len , file_index, output);

		memcpy(*p, tmp_m[0], entry_size*sizeof(double));
		m_free(tmp_m);

		if(esize != entry_size) PRINTF("entry_size changed !!!!");
	}
}

void DiskDataDesc::GetSamples (double** P, int* I, int len) {
	
	int* I_tmp    = int_array(len);
	int I_tmp_len = 0;

	for (int i=0;i < len; i++) { // check which samples already have been loaded
				 // I_tmp contains those samples that still have to be loaded
		if (!store->is_in_mem(I[i])) {
			I_tmp[I_tmp_len] = i;
			I_tmp_len++;
		}
		else P[i] = store->ref(I[i]);
	}

	if (I_tmp_len > 0) { // load the rest from file
		int* uinfo    = int_array(I_tmp_len);
		iget_i(uinfo, I, I_tmp, I_tmp_len);

		int uinfo_len = I_tmp_len;
		int reorder   = 1;
		int tmp       = 0;
		int esize     = entry_size;
		int output    = 0;

		double** tmp_m = load_data (filename, &entry_size, &tmp, reorder, uinfo, uinfo_len , file_index, output);
		free_array(uinfo);

		if(esize != entry_size) PRINTF("entry_size changed !!!!");

		for (int i=0; i < I_tmp_len; i++) {
			P[I_tmp[i]] = store->ref(I[I_tmp[i]]);
			memcpy(P[I[I_tmp[i]]], tmp_m[i], entry_size*sizeof(double));
		}

		m_free(tmp_m);
	}
	free_array(I_tmp);
}

int DiskDataDesc::CopySample (double* x, int** NZ, int i) {
	double* p = NULL;
	GetSample(&p, i);
	memcpy(x, p, entry_size*sizeof(double));
	
	if (NZ != NULL) *NZ = NULL;

	return entry_size;
}

void DiskDataDesc::PutSample(double* x, int* NZ, int NZ_len, int i) {
	PRINTF("DiskDataDesc: No write support!\n");
}

int DiskDataDesc::ssize (void) {
	return entry_size;
}

int DiskDataDesc::num_samples (void) {
	return file_index_len;
}

int DiskDataDesc::Count(void) {
	return file_index_len*entry_size;
}

void DiskDataDesc::Flush(void) {
	if(save_memory)	store->flush();
}

void DiskDataDesc::getAbsIndex (int* index) { 
	for (int i=0; i < file_index_len; i++) index[i] = i; 
}
void DiskDataDesc::getAbsIndex (int* index, int* I, int len) { 
	memcpy(index, I, len*sizeof(int)); 
}

///////////////////////////////////////////////////////////////////
/////////////////////////
////////////////////////////////////////////////////////////////
//

SubsetDataDesc::SubsetDataDesc (int* S_tmp, int S_len, DataDesc* D) : ProxyDataDesc(D), S_len(S_len) {
	S = int_array(S_len);
	memcpy(S, S_tmp, S_len*sizeof(int));
}

SubsetDataDesc::~SubsetDataDesc(void) {
	free_array(S);
}
	
void SubsetDataDesc::GetSample  (double** p, int i) {
	pdesc->GetSample(p, S[i]);
}

void SubsetDataDesc::GetSamples (double** P, int* I, int len) {
	int* itemp = int_array(len);

	iget_i(itemp, S, I, len);
	pdesc->GetSamples(P, itemp, len);

	free_array(itemp);
}

int SubsetDataDesc::CopySample (double* x, int** NZ, int i) {
	return pdesc->CopySample(x, NZ, S[i]);
}

void SubsetDataDesc::PutSample(double* x, int* NZ, int NZ_len, int i) {
	pdesc->PutSample(x, NZ, NZ_len, S[i]);
}

int SubsetDataDesc::num_samples (void) {
	return S_len;
}

void SubsetDataDesc::getAbsIndex (int* index) { 
	memcpy(index, S, S_len*sizeof(int)); 
}
void SubsetDataDesc::getAbsIndex (int* index, int* I, int len) {
	int* tmp_index = int_array(len);
	iget_i(tmp_index, S, I, len);
	
	ProxyDataDesc::getAbsIndex(index, tmp_index, len);

	free_array(tmp_index);
}

///////////////////////////////////////
//
///////////////////////////////////////

DimSelectDataDesc::DimSelectDataDesc(int* M_tmp, int M_len, DataDesc* desc, int save_mem) : ProxyDataDesc(desc), M_len(M_len), save_memory(save_mem) {

	M = int_array(M_len);
	memcpy(M, M_tmp, M_len*sizeof(int));

	store =  new MemStore(num_samples(), M_len);
}

DimSelectDataDesc::~DimSelectDataDesc (void) {
	free_array(M);
	
	delete store;
}

void DimSelectDataDesc::GetSample (double** p, int i) {


	if (store->is_in_mem(i)) *p = store->ref(i);
	else {
		double* p_tmp = NULL;
		ProxyDataDesc::GetSample(&p_tmp, i);
		*p = store->ref(i);
		get_i(*p, p_tmp, M, M_len);
		ProxyDataDesc::Flush();
	}
}

void DimSelectDataDesc::GetSamples (double** P, int* I, int len) {
	for (int i=0; i < len; i++)  GetSample(P + i, I[i]);
}

int DimSelectDataDesc::CopySample (double* x, int** NZ, int i) {

	if (store->is_in_mem(i)) {
		 double* tmp = store->ref(i);
		 memcpy(x,tmp, ssize()*sizeof(double));
	}
	else {
		double* p_tmp = NULL;
		ProxyDataDesc::GetSample(&p_tmp, i);
		double* tmp = store->ref(i);
		get_i(tmp, p_tmp, M, M_len);
		ProxyDataDesc::Flush();

		memcpy(x, tmp, ssize()*sizeof(double));
	}
	
	if (NZ != NULL) *NZ =NULL;

	return ssize();
}

void DimSelectDataDesc::GetDimIndex (int* dest, int* dims, int len) {
	iassign(dest, -1, len);
	for (int i=0; i < len; i++) ifind_eq(dest + i, M, dims[i], ssize());
}

void DimSelectDataDesc::GetIndex(int* dest) {
	memcpy(dest, M, ssize()*sizeof(int));
}

void DimSelectDataDesc::remove_dims (int* d, int count) {

	for (int j=0;j < count; j++) {

		int pos=0;

		ifind_eq(&pos, M, d[j], ssize());
		for (int i=pos+1; i < ssize(); i++) M[i-1]=M[i];

		for (int i=0; i < size(); i++) {
			if (store->is_in_mem(i)) {
				double* tmp = store->ref(i);
				double* limit = tmp + ssize();

				tmp = tmp + pos + 1;

				while(tmp < limit) *(tmp-1) = *tmp++; 
			}
		}

		M_len -= 1;

		store->set_entry_size(M_len);
	}
}

void DimSelectDataDesc::PutSample(double* x, int* NZ, int NZ_len, int i) {
	PRINTF("DimSelectDataDesk: No write support !!!\n");
}

int DimSelectDataDesc::ssize(void) {
	return M_len;
}

void DimSelectDataDesc::Flush(void) {
//	PRINTF("DimSelectDataDesc::Flush, save_memory: %i\n", save_memory);

	ProxyDataDesc::Flush();
	if (save_memory) store->flush();
}

///////////////////////////////////////
//
///////////////////////////////////////

RemoveNANsDataDesc::RemoveNANsDataDesc(DataDesc* desc) : ProxyDataDesc(desc), valid_dims(NULL) {

	valid_dims = int_array(ssize());
}

RemoveNANsDataDesc::~RemoveNANsDataDesc (void) {
	free_array(valid_dims);
}

int RemoveNANsDataDesc::CopySample (double* x, int** NZ, int i) {
	
	double* p_tmp = NULL;
	ProxyDataDesc::GetSample(&p_tmp, i);

	memcpy(x, p_tmp, ssize()*sizeof(double));

	int ret = isAN(valid_dims, x, ssize());

	if (NZ != NULL) {
		if (ret != ssize()) *NZ = valid_dims;	
		else *NZ = NULL;
	}

	return ret;
}

//////////////////////////////////////////////////77
///
/////////////////////////////////////

ClassInfoDataDesc::ClassInfoDataDesc(DataDesc* desc) : ProxyDataDesc(desc), c(NULL) {
}

ClassInfoDataDesc::ClassInfoDataDesc(char* data_file, char* class_info_file, int save_memory) : c(NULL), ProxyDataDesc(new DiskDataDesc(data_file, save_memory)) {

	int reorganize= 1;
	int c_rows    = 0;
	int c_columns = 0;

	double* file_index=NULL; 
 	int output = 0;
	
	double** c_temp = load_data(class_info_file, &c_rows, &c_columns, reorganize, NULL,0, file_index, output);

	if (!c_temp) {
		printf("Could not load label file: %s\n", class_info_file);
	}
	else {
		c = double_array(c_rows*c_columns);
		m_to_vec(c, c_temp, c_columns, c_rows); 
		m_free(c_temp);
	}
}

ClassInfoDataDesc::ClassInfoDataDesc(double* ct, int ctlen, DataDesc* desc) : ProxyDataDesc(desc) {
	if (ctlen != desc->num_samples()) PRINTF("ClassInfoDataDesc size of ct does not equal number of samples in desc !!!!\n");

	c = double_array(desc->num_samples());
	memcpy(c, ct, desc->num_samples()*sizeof(double));
}

ClassInfoDataDesc::~ClassInfoDataDesc(void) {
	free_array(c);
	delete pdesc;
}

double ClassInfoDataDesc::getClassInfo(int i) {
	return c[i];
}

void ClassInfoDataDesc::getClassInfo(double* c_dest) {
	memcpy(c_dest, c, pdesc->size()*sizeof(double));
}

void ClassInfoDataDesc::getClassInfo(double* c_dest, int* I, int I_len) {
	get_i(c_dest, c, I, I_len);
}

/////////////////////////////////////////////////////
///
/////////////////////////////////////////////////////

ClassInfoDimSelectDataDesc::ClassInfoDimSelectDataDesc(int* M_tmp, int M_len, ClassInfoDataDesc* desc) : ClassInfoDataDesc (new DimSelectDataDesc(M_tmp, M_len, desc)) {

	c = double_array(num_samples());
	desc->getClassInfo(c);

	dim_select = dynamic_cast<DimSelectDataDesc*>(pdesc);
}

void ClassInfoDimSelectDataDesc::remove_dims (int* d, int count) {
	dim_select->remove_dims(d, count);
}

void ClassInfoDimSelectDataDesc::GetDimIndex (int* dest, int* dims, int len) {
	dim_select->GetDimIndex(dest, dims, len);
}

void ClassInfoDimSelectDataDesc::GetIndex(int* dest) {
	dim_select->GetIndex(dest);
}

////////////////////////////////////////////////////
///
//////////////////////////////////////////////////

SubsetClassInfoDataDesc::SubsetClassInfoDataDesc(int* S_tmp, int S_len, ClassInfoDataDesc* D) : ClassInfoDataDesc(new SubsetDataDesc(S_tmp, S_len, D)) {
	double* c_temp = double_array(D->num_samples());

	D->getClassInfo(c_temp);

	c = double_array(num_samples());
	get_i(c, c_temp, S_tmp, S_len);

	free_array(c_temp);
}

SubsetClassInfoDataDesc::~SubsetClassInfoDataDesc(void) {
}

void SubsetClassInfoDataDesc::getClassInfo(double* c_dest) {
	memcpy(c_dest, c, pdesc->size()*sizeof(double));
}

double SubsetClassInfoDataDesc::getClassInfo(int i) {
	return c[i];
}

void SubsetClassInfoDataDesc::getClassInfo(double* c_dest, int* I, int I_len) {
	get_i(c_dest, c, I, I_len);
}
