/*
written by Kai Labusch
Copyright (C) 2006 Kai Labusch

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Contact:
labusch@inb.uni-luebeck.de
*/

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <regex.h>
#include <cmath>

using namespace std;

#include "DiskUtils.h"
#include "utils.h"

int exist(char* filename) {
	struct stat buffer;
	int status = stat(filename, &buffer);	

	return (status == 0);
}

/*
	returns *columns entrys of size *rows
*/
double** load_data (char* filename, int* entry_size, int* num_entrys, int reorganize, int* use_info, int use_info_size,
		    double* file_index, int output) {

	if (output) PRINTF("Loading data from %s ...", filename);

	double** M = NULL;
	if (parse_data (&M, entry_size, num_entrys, filename, reorganize, use_info, use_info_size, file_index)) {
		if (output) PRINTF("finished.\n");
	}
	else  printf ("Could not open: %s \n", filename);

	return M;
}

// save each !column! of matrix M in each row of file "filename"
int save_data (char* filename, double** M, int entry_size, int nentrys) {
	FILE* foutput = fopen (filename, "w");

	if (foutput != NULL) {
		int i=0;
		int j=0;	
		
		for (j=0; j < nentrys; j++) {
			for (i=0; i < entry_size; i++) {
				if ( ((double) ((int)M[j][i])) - M[j][i] != 0.0) fprintf (foutput, "%g ", M[j][i]);
				else fprintf (foutput, "%i ", (int) M[j][i]);
			}

			fprintf (foutput, "\n");
		}

		fclose(foutput);

		return 1;
	}

	printf ("Could not open: %s \n", filename);

	return 0;
}

int append_data (const char *filename, double *M, int columns) {
    FILE *foutput = fopen (filename, "a");

    if (foutput != NULL) {
        int i=0;

        for (i=0; i<columns; i++)
            fprintf (foutput, "%.8lg ", M[i]);

        fprintf (foutput, "\n");
        fclose(foutput);

        return 1;
    }

    printf ("Could not open: %s \n", filename);

    return 0;
}

int save_int_data (char* filename, int** M, int rows, int columns) {
	FILE* foutput = fopen (filename, "w");

	if (foutput != NULL) {
		int i=0;
		int j=0;	
		
		for (j=0; j < columns; j++) {
			for (i=0; i < rows; i++) fprintf (foutput, "%i ", (M[j])[i]);

			fprintf (foutput, "\n");
		}

		fclose(foutput);

		return 1;
	}

	printf ("Could not open: %s \n", filename);

	return 0;
}

double* parse_string (char* str, int* len) {
	char* pos = str;
	char* new_pos = pos;

	*len=0;

	do {
		pos = new_pos;	
		double tmp = strtod(pos, &new_pos);

		if (new_pos != pos) (*len)++;

	} while (new_pos != pos);
	double* ret = double_array(*len);
	memset (ret, 0, *len * sizeof(double)); 
	double* retp = ret;

	pos = new_pos = str;
	do {
		pos = new_pos;
		*(retp++) = strtod(pos, &new_pos);
	} while (new_pos != pos);
	return ret;
}

int parse_line (double* line, char* buf) {

	char* new_pos=buf;
	char* row_pos=buf;
	int rcount = 0;
	do {
		row_pos = new_pos;
		double d = strtod(row_pos, &new_pos);

		if ((line != NULL) && (new_pos != row_pos)) line[rcount] = d;

		if (new_pos != row_pos) rcount++;

	} while (new_pos != row_pos);
	
	return rcount;
}

double* create_file_index (char* filename, int* index_length) {

	int buf_size = 10*1024*1024;
	double* temp_index = double_array(buf_size);	
	char* buf    = char_array (buf_size);
	FILE* finput = fopen (filename, "r");
	double* ret = NULL;
	*index_length=0;

	if (finput != NULL) {
		PRINTF("create_file_index: %s ...", filename);

		int index_pos = 0;
		temp_index[index_pos] = ftello(finput);

		while (fgets(buf, buf_size, finput)) {
			index_pos++;
			temp_index[index_pos] = ftello(finput);
		}

		*index_length = index_pos;
		ret = double_array(index_pos);
		memcpy(ret, temp_index, index_pos*sizeof(double));

		fclose(finput);

		PRINTF(" length: %i.\n", *index_length);
	}
	else PRINTF("create_file_index: error. could not open file: %s\n", filename);

	free_array(temp_index);
	free_array(buf);

	return ret;
}

int parse_data (double*** M, int* entry_size, int* num_entrys, char* filename, int reorganize, int*use_info, int use_info_size, double* file_index) {

	FILE* finput = fopen (filename, "r");

	if (finput != NULL) {

		int buf_size   = 10*1024*1024;
		char* buf      = char_array (buf_size);
		int p_buf_size = 1000000;
		double** p_buf = (double**) pointer_array(p_buf_size);
		memset(p_buf, 0, p_buf_size*sizeof(double**));

		*entry_size= -1;
		*num_entrys= 0;
		int use_row = 0;
		int file_pos = 0;
		int* use_info_endpos = NULL;

		int* temp_use_info = NULL;
		int* reorder = NULL;
		int i=0;
		
		if (use_info) {
			temp_use_info = int_array(use_info_size);
			reorder       = int_array(use_info_size);
			for(i=0;i < use_info_size;i++) reorder[i] = i;

			memcpy(temp_use_info, use_info, use_info_size*sizeof(int));
			i_isort(temp_use_info, reorder, use_info_size); 

			use_info_endpos = temp_use_info + use_info_size-1;
		}
		int* cur_use_info = temp_use_info;

		if ((file_index == NULL) || (use_info == NULL)){
			while (fgets(buf, buf_size, finput)) {

				if (use_info) while((*cur_use_info < file_pos) &&(cur_use_info != use_info_endpos))cur_use_info++;

				use_row = (temp_use_info != NULL) ? (*cur_use_info == file_pos) : 1;

				if (use_row) {

					if (*entry_size == -1) *entry_size = parse_line (NULL, buf);
		
					if (p_buf[*num_entrys] == NULL) p_buf[*num_entrys] = double_array(*entry_size);
					parse_line(p_buf[*num_entrys], buf);
					(*num_entrys)++;
				}
				file_pos++;
			}
		}
		else {

			for (i = 0; i < use_info_size; i++) {
				fseeko(finput, (off_t) (file_index[temp_use_info[i]]), SEEK_SET);
				char* tmp = fgets(buf, buf_size, finput);
				
				if (*entry_size== -1) *entry_size= parse_line (NULL, buf);
		
				if (p_buf[*num_entrys] == NULL) p_buf[*num_entrys] = double_array(*entry_size);
				parse_line(p_buf[*num_entrys], buf);
				(*num_entrys)++;
			}
		}

		if (reorder) {
			double** tmp_p = p_buf;
			p_buf = (double**) pointer_array(p_buf_size);
			pset_i(p_buf, tmp_p, reorder, use_info_size);
			free_array(reorder);
			free_array(temp_use_info);
			free_array(tmp_p);
		}
		fclose(finput);

		if (reorganize) {
			*M = m_create(*num_entrys, *entry_size);
			m_copy(*M, p_buf, *num_entrys, *entry_size);
	
			int i=0;
			for(i=0; i < *num_entrys; i++) free_array(p_buf[i]);
			free_array(p_buf);
		}
		else *M = p_buf;
		free_array(buf);
		return 1;
	}
	return 0;
}

double** split_data (double** D, int D_rows, int D_columns, double** c) {

	double** X = m_create(D_columns, D_rows - 1);
	*c = double_array(D_columns);

	int X_rows = D_rows - 1;

	int i=0;
	for (i=0; i < D_columns; i++) {
		memcpy(X[i], D[i], X_rows *sizeof(double));
		(*c)[i] = (D[i])[X_rows];
	}

	return X;
}


char** get_files (char* dirname, int* num_files) {
	
	DIR* directory = opendir(dirname);
	*num_files = 0;

	if (directory != NULL) {
		struct dirent* centr;
		int cur_size = 10;
		char** result = (char**) pointer_array(cur_size); 
	
		while ((centr = readdir(directory)) != NULL) {
			if (*num_files >= cur_size) {
				char** old = result;
				result = (char**) pointer_array(cur_size + 100); 
				memcpy (result, old, cur_size * sizeof(char**));
				free_array(old);
				cur_size += 100;
			}

			result[*num_files] = char_array(strlen(centr->d_name) + 1);	
			sprintf(result[*num_files], "%s", centr->d_name);
	
			(*num_files)++; 
		}

		closedir(directory);

		return result;
	}

	return NULL;
}

int match(const char *string, char *pattern) {
	int    status;
	regex_t    re;

	if (regcomp(&re, pattern, REG_EXTENDED|REG_NOSUB) != 0) return 0;

	status = regexec(&re, string, (size_t) 0, NULL, 0);
	regfree(&re);

	return !status;
}

int compare_files (const void* a, const void *b) {
	char* ca = *((char**) a);
	char* cb = *((char**) b);

	return strcmp(ca, cb);
}

char** str_sort (char** str, int len) {

	printf ("len: %i\n", len);
	char** ret = (char**) pointer_array(len);
	memcpy(ret, str, len*sizeof(char**));	

	qsort((void*)ret, len, sizeof(char**), compare_files);
	
	return ret;
}

char** filter (char** str, int len, char* pattern, int* new_len) {

	char** ret = (char**) pointer_array(len);
	memset(ret, 0, len * sizeof(char**));

	int i=0;
	*new_len = 0;

	for (i=0; i < len; i++) 
		if (match(str[i], pattern)) ret[(*new_len)++] = str[i]; 

	return ret;
}

int get_bin_file_size(char* filename, int* num_rows, int* num_columns, int* clength, int* entry_size) {
	FILE* finput = fopen (filename, "r");

	if (!finput) return 0;

	char buf[12];
	size_t tmp = fread(buf, 1, 12, finput);

	if (!is_little_endian()) toogle_endianess((int*) buf, 3);

	*num_rows    = *(((int*) buf) + 0);
	*num_columns = *(((int*) buf) + 1);
	*entry_size  = *(((int*) buf) + 2);

	*clength = (*num_rows * *entry_size + 7)/8;/* ceil */

	fclose(finput);

	return 1;
}

int get_bytes(char* dest, FILE* file, int clen, int tcolumn, int offset, int bytes_read) {
	if (!fseeko(file, ((off_t)tcolumn)*((off_t)clen) + (off_t)12 + ((off_t) offset), SEEK_SET)) {
		if (fread(dest, 1, bytes_read, file) != bytes_read) {
			memset(dest, 0, bytes_read);
			return 0;
		}
		else return 1;
	}
	else {
		memset(dest, 0, bytes_read);
		return 0;
	}
}

int put_bytes(FILE* file, char* src, int clen, int tcolumn, int offset, int bytes_write) {
	
	off_t goffset = ((off_t)tcolumn)*((off_t)clen) + ((off_t)12) + ((off_t)offset);

	if (!fseeko(file, goffset, SEEK_SET)) {
		if (fwrite(src, 1, bytes_write, file) != bytes_write) {
			PRINTF("error during write operation on file!!!!");
			return 0;
		}	
		return 1;	
	}
	else {
		fseeko(file, ((off_t)0), SEEK_END);
		off_t pos  = ftello(file);
		off_t fill = ((off_t)tcolumn + 1) - (pos - ((off_t)12))/((off_t)clen); 
		off_t i=0;
		char* buf = char_array(clen);
		memset(buf,0, clen);
		for (i=0; i < fill; i++) size_t tmp=fwrite(buf, 1, clen, file);
		free_array(buf);

		return put_bytes(file, src, clen, tcolumn, offset, bytes_write);
	}
}
