#ifndef __MXDATA_H_INCLUDED__
#define __MXDATA_H_INCLUDED__

#include <cstring>

#include "mex.h" 

class MXDataDesc : public DataDesc {
public:

	static DataDesc* Create(const mxArray* mx_data, int save_memory);

	static ClassInfoDataDesc* Create(const mxArray* mx_data, const mxArray* mx_c, int save_memory);

	static SubsetDataDesc* Create(DataDesc* desc, const mxArray* subset_index);

	MXDataDesc(const mxArray* mx_data) : DataDesc(), entry_size(mxGetM(mx_data)), size(mxGetN(mx_data)),mx_data(mx_data) {
	}

	~MXDataDesc(void) { }

	virtual int ssize(void) { return entry_size; }
	virtual int num_samples (void) { return size; }	
	virtual int Count(void) { return size*entry_size; }

	virtual void GetSample(double** x, int i)=0;
	virtual void GetSamples (double** P, int* I, int len)=0;

	virtual int CopySample (double* x, int** NZ, int i)=0;
	virtual void PutSample(double* x, int* NZ, int NZ_len, int i)=0;

	virtual void Flush(void)=0;

	virtual void getAbsIndex (int* index) { for (int i=0; i < size; i++) index[i] = i; }
	virtual void getAbsIndex (int* index, int* I, int len) { memcpy(index, I, len*sizeof(int)); }

	virtual DataDesc* ID(void) { return this; }

protected:
	const mxArray* mx_data;

	int size;
	int entry_size;
};


class MXUINT8DataDesc: public MXDataDesc {
public:
	MXUINT8DataDesc(const mxArray* mx_data, int save_memory);
	~MXUINT8DataDesc(void);

	virtual void GetSample(double** x, int i);
	virtual void GetSamples (double** P, int* I, int len);

	virtual int CopySample (double* x, int** NZ, int i);
	virtual void PutSample(double* x, int* NZ, int NZ_len, int i);

	virtual void Flush(void);
private:
	unsigned char** D;

	MemStore* store;

	int save_memory;
};

class MXINT32DataDesc: public MXDataDesc {
public:
	MXINT32DataDesc(const mxArray* mx_data, int save_memory);
	~MXINT32DataDesc(void);

	virtual void GetSample(double** x, int i);
	virtual void GetSamples (double** P, int* I, int len);

	virtual int CopySample (double* x, int** NZ, int i);
	virtual void PutSample(double* x, int* NZ, int NZ_len, int i);

	virtual void Flush(void);
private:
	int** D;

	MemStore* store;

	int save_memory;
};

#endif
