#include <cstdio>
#include <cmath>
#include <cstdlib>
#include <cstring>

#include "utils.h"
#include "min.h"
#include "knn.h"

double mse (double** X, int X_columns, int ssize, double** codebook, int codebook_size) {

	double* temp = double_array(X_columns);
	double* temp2 = double_array(codebook_size);

	int i=0;
	int j=0;
	for (i=0; i < X_columns; i++) {
		for (j=0; j < codebook_size; j++) temp2[j] = diff_scp(X[i], codebook[j], ssize);

		temp[i] = temp2[dmin(temp2, codebook_size)];	
	}

	double rmse = mean(temp, X_columns);

	free_array(temp);
	free_array(temp2);

	return rmse;
}

double** kmeans(double** X, int X_columns, int ssize, int codebook_size, double accuracy, double** r_labels) {

	int i=0;
	int j=0;

	double** codebook 	= m_create(codebook_size, ssize);
	double** tmp_codebook   = m_create(codebook_size, ssize);
	double* counter 	= double_array(codebook_size);
	*r_labels = double_array(X_columns);

	/* init codebook */
	double* vmin = double_array(ssize);
	double* vmax = double_array(ssize);

	memcpy(vmin, X[0], sizeof(double)*ssize);
	memcpy(vmax, X[0], sizeof(double)*ssize);
	for (i=0; i < X_columns; i++) {
		for (j=0; j < ssize; j++) {
			vmin[j] = (vmin[j] < X[i][j]) ? vmin[j] : X[i][j];
			vmax[j] = (vmax[j] > X[i][j]) ? vmax[j] : X[i][j];	
		}
	}

	double* vtemp = double_array(codebook_size);
	for (j=0; j < ssize; j++) {
		drand_array(vtemp, vmin[j], vmax[j], codebook_size);	
		for (i=0; i < codebook_size; i++) codebook[i][j] = vtemp[i];
	}
	free_array(vmin);	
	free_array(vmax);
	free_array(vtemp);

	double* norm = double_array(codebook_size);
	int* index   = int_array(codebook_size);

	int left  = 0;
	int right = 0;
	int stop  = 0;

	int k = 1;
	double* heap   = double_array(k);
	int*    heap_I = int_array(k);
	int heap_size  = 0;
	double cur_norm = 0;

	int acc_reached = 0;
	while (!acc_reached) {

		/* calculate norm of each codebook vector*/
		for (i=0;i < codebook_size; i++) {
			norm[i] = sqrt(scp(codebook[i],codebook[i], ssize));
			index[i]= i;
		}

		/* sort norm of each codebook vector*/
		isort(norm, index, codebook_size);

		m_zero(tmp_codebook,  codebook_size, ssize);
		memset(counter, 0, codebook_size*sizeof(double));	

		for (i=0;i < X_columns;i++) { /* find closest codebook vector for each sample */

			memset(heap    , 0, k*sizeof(double)); /* reset heap */
			memset(heap_I  , 0, k*sizeof(int));
	
			heap_size = 0;	
			left = right = -1;
			stop  = 0;
			cur_norm = sqrt(scp(X[i],X[i],ssize));

			while (!stop) stop = insert_knn_sample (codebook, X[i], ssize, heap, heap_I, &heap_size, k, codebook_size,
			      					norm, index, cur_norm, &left, &right);
			counter[heap_I[0]]++;
			add(tmp_codebook[heap_I[0]], tmp_codebook[heap_I[0]], X[i], ssize);

			(*r_labels)[i] = heap_I[0];
		}

		acc_reached = 1;
		for (i=0; i < codebook_size; i++) { /* update codebook vectors */

			if (counter[i] > 0) {

				sdivide(tmp_codebook[i], tmp_codebook[i], counter[i], ssize);

				double distance = sqrt(diff_scp(tmp_codebook[i], codebook[i], ssize));

				if (distance > accuracy) {
					acc_reached = 0;
					memcpy(codebook[i], tmp_codebook[i], ssize*sizeof(double));
				}
			}
		}
	}

	free_array(norm);
	free_array(index);
	free_array(heap);
	free_array(heap_I);
	
	m_free(tmp_codebook);
	free_array(counter);

	return codebook;
}
