/*
(Soft)MaxMinOver learning algorithm written by Kai Labusch
Copyright (C) 2004 Kai Labusch

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Contact:
labusch@inb.uni-luebeck.de
*/


#include "min.h"
#include "utils.h"

#include <cmath>
#include <cstring>

using namespace std;

int kmin (double* Y, int* I, double* X, int x_size, int k) {
	int used = 0;
	int i;

	memset (Y, 0, k*sizeof(double));
	
	for (i=0; i < x_size; i++) {

		if (used < k) {
			used = heap_insert (Y,I, X[i], i, used);
		}
		else if (X[i] < Y[0]) {
			used = heap_extract_max (Y, I, used);
			used = heap_insert (Y,I, X[i], i, used);
		}
	}

	return used;
}

void heapify (double* H, int* I, int index, int used) {
	int l = index << 1;
	int r = l + 1;
	int largest = 0;

	if ((l <= used) && (H[l-1] > H[index-1])) {
		largest = l;
	}
	else {
		largest = index;
	}

	if ((r <= used) && (H[r-1] > H[largest-1])) {
		largest = r;
	}

	if (largest != index) {
		double temp =  H [largest-1];
		H[largest-1] = H[index-1];
		H[index-1] = temp;

		int itemp = I[largest-1];
		I[largest-1] = I[index-1];
		I[index-1] = itemp;

		heapify (H, I, largest, used);
	}
}

int heap_extract_max (double* H, int* I, int used) {
	int n_used = used;

	if (used < 1) return n_used;

	H[0] = H[used-1];
	I[0] = I[used-1];

	n_used = used -1;

	heapify (H, I, 1, n_used);

	return n_used;
}

int heap_insert (double* H, int* I, double x, int x_I, int used) {
	used = used + 1;
	int i = used;

	int parent = i >> 1;

	while ((i > 1) && (H[parent - 1] < x)) {
		H[i-1] = H[parent-1];
		I[i-1] = I[parent-1];

		i = parent;
		parent = i >> 1;
	}

	H[i-1] = x;
	I[i-1] = x_I;

	return used;
}



