/*
neural gas algorithm written by Kai Labusch
Copyright (C) 2007 Kai Labusch

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Contact:
labusch@inb.uni-luebeck.de
*/


#include <cstdio>
#include <cmath>
#include <cstdlib>
#include <cstring>

using namespace std;

#include "utils.h"

double** ngas(double** X, int X_columns, int ssize, int codebook_size, int epochs, double alpha0, double lambda0, double** r_labels) {

	double** codebook = m_create(codebook_size, ssize);

	for(int i=0; i < codebook_size; i++) drand_array(codebook[i], -0.5*10e-5, 0.5*10e-5, ssize);

	int train_steps = epochs*X_columns;

	if (alpha0 == 0.0)  alpha0  = 0.5;
	if (lambda0 == 0.0) lambda0 = ((double) codebook_size)/2.0;

	double alphaf  = 0.005;
	double lambdaf = 0.01; 

	double* temp = double_array(codebook_size);
	int* itemp = int_array(codebook_size);

	double* ranking = double_array(codebook_size);
	double* h = double_array(codebook_size);

	double* cdiff = double_array(ssize);	

	double old_percent = 0.0;
	for (int i=0; i < train_steps; i++) {
		double percent = floor( ((double)i)*100/train_steps );
		if (old_percent != percent) {
			old_percent = percent;
			PRINTF("%g %%\n", old_percent);
			FLUSH;
		}
	
		// compute current learning rate as well as current neighbourhood size
		double lambda = lambda0 * pow(lambdaf/lambda0, ((double)i)/((double) train_steps));
		double alpha  = alpha0  * pow(alphaf/alpha0  , ((double)i)/((double) train_steps));

		// select random sample for learning
		double* x = X[(int) (((X_columns - 1)*(double)rand())/(RAND_MAX + 1.0))];

		// calculate distance of current sample to all codebook vectors
		for (int j=0; j < codebook_size; j++) {
			temp[j]  = diff_scp(x, codebook[j], ssize);
			itemp[j] = j; 
		}

		// calculate neighbourhood ranking
		isort(temp, itemp, codebook_size);	
		linspace(temp, 0, codebook_size - 1, codebook_size);
		set_i(ranking, temp, itemp, codebook_size);
		smultiply(ranking, ranking, -1.0/lambda, codebook_size);	
		exponential(h, ranking, codebook_size); 

		// update codebook vectors
		for (int j=0; j < codebook_size; j++) {
			sub(cdiff, x, codebook[j], ssize); 	
			smultiply(cdiff, cdiff, alpha*h[j], ssize);
			add(codebook[j], cdiff, codebook[j], ssize);
		}	
	}

	// determine codebook labels
	*r_labels = double_array(X_columns);
	for (int i=0; i < X_columns; i++) {

		// calculate distance of current sample to all codebook vectors
		for (int j=0; j < codebook_size; j++) temp[j]  = diff_scp(X[i], codebook[j], ssize);
		
		(*r_labels)[i] = dmin(temp, codebook_size);
	}

	free_array(temp);
	free_array(itemp);
	free_array(ranking);
	free_array(h);
	free_array(cdiff);

	return codebook;
}
