/*
Some helper functions written by Kai Labusch
Copyright (C) 2004-2010 Kai Labusch

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Contact:
labusch@inb.uni-luebeck.de
*/

#ifndef __UTILS_H_INCLUDED__
#define __UTILS_H_INCLUDED__

/*#define __USE_SSE__*/

/*-------------------------------------------*/
/* to be modified if the source is to be used in a non MEX environment*/

#ifdef __MATLAB_USAGE__ 

#include "mex.h" 

//#define MALLOC mxMalloc  // do not use MATLAB memory allocation !!! it is broken!
//#define FREE   mxFree
#define MALLOC malloc
#define FREE free
#define PRINTF mexPrintf
#define FLUSH  mexEvalString("drawnow;")

#else

#ifdef __PARALLEL_USAGE__

#define MALLOC malloc
#define FREE free
#define PRINTF empty_printf
#define FLUSH

#else

#include <stdio.h>

#define MALLOC malloc
#define FREE free
#define PRINTF printf
#define FLUSH

void dbgconsole (const char* format, ...);

//#define PRINTF dbgconsole
//#define PRINTF

#endif


#endif

#ifdef MTMALLOC
#include <mtmalloc.h>
#endif

#ifndef _MINDEX64

typedef int MIndex;

#else

typedef long MIndex;

#endif


void print_array (const char* text, double* array, int length);
void iprint_array (const char* text, int* array, int length);
/*---------*/

void assign (double* dest, double val, int length);
void assign_i(double* dest, int* I, double val, int length);
/*---------*/

void iassign (int* dest, int val, int length);
void iassign_i(int* dest, int* I, int val, int length);
/*---------*/

void multiply (double* product, double* factor1, double* factor2, int length);
void multiply_ip (double* factor1, double* factor2, int length);
void multiply_i(double* product, double* factor1, double* factor2, int* I, int length);
void multiply_ip_i(double* factor1, double* factor2, int* I, int length);
void multiply_shr(double* product, double* factor1, double* factor2, int* I, int length);
/*----------*/

void divide (double* quotient, double* numerator, double* denominator, int length);
void divide_ip (double* numerator, double* denominator, int length);
void divide_i(double* quotient, double* numerator, double* denominator, int* I, int length);
void divide_ip_i(double* numerator, double* denominator, int* I, int length);
void divide_shr(double* quotient, double* numerator, double* denominator, int* I, int length);
/*---------*/

void add (double* sum, double* s1, double* s2, int length);
void add_ip (double* s1, double* s2, int length);
void add_i(double* sum, double* s1, double* s2, int* I, int length);
void add_ip_i(double* s1, double* s2, int* I, int length);
void add_shr(double* sum, double* s1, double* s2, int* I, int length);
/*---------*/


void sub (double* diff, double* sum1, double* sum2, int length);
void sub_ip (double* sum1, double* sum2, int length);
void sub_i(double* diff, double* sum1, double* sum2, int* I, int length);
void sub_ip_i(double* sum1, double* sum2, int* I, int length);
void sub_shr(double* diff, double* sum1, double* sum2, int* I, int length);
/*-------*/

void isub      (int* diff, int* sum1, int* sum2, int length);
void isub_ip   (int* sum1, int* sum2           , int length);
void isub_i    (int* diff, int* sum1, int* sum2, int* I, int length);
void isub_ip_i (int* sum1, int* sum2, int* I, int length);
void isub_shr  (int* diff, int* sum1, int* sum2, int* I, int length);
/*--------*/

void absol (double* xabs, double* x, int length);
void absol_i(double* xabs, double* x, int* I, int length);
/*-----------*/

void smultiply(double* product, double* fact1, double fac2, int length);
void smultiply_ip(double* fac1, double fac2, int length);
void smultiply_i(double* product, double* fact1, double fac2, int* I, int length);
void smultiply_ip_i(double* fact1, double fac2, int* I, int length);
void smultiply_shr(double* product, double* fact1, double fac2, int* I, int length);
/*------------*/

void sdivide(double* quotient, double* numerator, double denominator, int length);
void sdivide_ip(double* numerator, double denominator, int length);
void sdivide_i(double* quotient, double* numerator, double denominator, int* I, int length);
void sdivide_ip_i(double* numerator, double denominator, int* I, int length);
void sdivide_shr(double* quotient, double* numerator, double denominator, int* I, int length);
/*------------*/

void sadd(double* sum, double* s1, double s2, int length);
void isadd(int* sum, int* s1, int s2, int length);
void sadd_ip(double* s1, double s2, int length);
void sadd_i(double* sum, double* s1, double s2, int* I, int length);
void isadd_i(int* sum, int* s1, int s2, int* I, int length);
void sadd_ip_i(double* s1, double s2, int* I, int length);
void isadd_ip_i(int* s1, int s2, int* I, int length);

void psadd(double** sum, double** s1, int offset, int length);

void sadd_shr(double* sum, double* s1, double s2, int* I, int length);
/*-------------*/

void ssub(double* diff, double* sum1, double sum2, int length);
void issub(int* diff, int* sum1, int sum2, int length);
void ssub_ip(double* sum1, double sum2, int length);
void ssub_i(double* diff, double* sum1, double sum2, int* I, int length);
void ssub_ip_i(double* sum1, double sum2, int* I, int length);
void ssub_shr(double* diff, double* sum1, double sum2, int* I, int length);

void issub_ip(int* sum1, int sum2, int length);
/*--------------*/

void exponential (double* dest, double* x, int length);
void exponential_i(double* dest, double* x, int* I, int length);

void exponential10 (double* dest, double* x, int length);
/*--------------*/

void power (double* dest, double* x, double y, int length);
void power2 (double* dest, double x, double* y, int length);
void power_i(double* dest, double* x, double y, int* I, int length);
/*---------------*/

void squarert (double* dest, double* x, int length);
void squarert_i (double* dest, double* x, int* I, int length);
/*---------------*/

double mean (double* x, int length);
double mean_i (double* x, int* I, int length);

double var (double* x, int length);
/*---------------*/

double prod(double* x, int length);
int prod_i(int* x, int length);

double sum (double* x, int length);
double sum_i (double* x, int* I, int length);
/*---------------*/

int isum (int* x, int length);
int isum_i (int* x, int* I, int length);
/*---------------*/

void rand_array (int* dest, int min, int max, int dest_length);
void drand_array (double* dest, double min, double max, int dest_length);

int uv_hash_func (int* hash_func_params, int* I, int I_length, int max);
/*----------*/
int dmax (double* x, int length);
int imax (int* x, int length);

/*----------*/
int dmin (double* x, int length);
int imin (int* x, int length);
int imin_i (int* x, int* I, int length);
/*----------------------------------------------------------------------------*/

int unique(double* un, double* x, int length, int* I=NULL);
int iunique(int* un, int* x, int length);
void isort(double* x, int* index, int length);
void i_isort(int* x, int* index, int length);

int dcompare (const void* pa, const void* pb);
int icompare (const void* pa, const void* pb);

int get_urandom_entries (int* dest, int* rest, int* I, int I_length, int count);
void randperm(int* p, int len);

/*----------------------------------------------------------------------------*/

int isNAN(int* dest, double* x, int length); // is not a number
int isAN(int* dest, double* x, int length);  // is a number

int find_neq (int* dest_indizes, double* x, double val, int x_length);
int ifind_neq (int* dest_indizes, int* x, int val, int x_length);
int cfind_neq (int* dest_indizes, char* x, char val, int x_length);
int find_eq  (int* dest_indizes, double* x, double val, int x_length);
int ifind_eq  (int* dest_indizes, int* x, int val, int x_length);
int sfind_eq  (int* dest_indizes, char** strs, char* str, int strs_length);
int sfind_eq  (int* dest_indizes, char** strs, const char* str, int strs_length);
int find_l (int* dest_indizes, double* x, double val, int x_length);
int find_leq (int* dest_indizes, double* x, double val, int x_length);
int find_g (int* dest_indizes, double* x, double val, int x_length);
int find_geq (int* dest_indizes, double* x, double val, int x_length);

int get_i(double* dest, double* x, int* indizes, int ind_length);
void set_i(double* dest, double* x, int* indizes, int ind_length);
void copy_i(double* dest, double* x, int* dest_indizes, int* x_indizes, int ind_length);

int iget_i(int* dest, int* x, int* indizes, int ind_length);
void iset_i(int* dest, int* x, int* indizes, int ind_length);
int cget_i(char* dest, char* x, int* indizes, int ind_length);
void cset_i(char* dest, char* x, int* indizes, int ind_length);
int pget_i(double** dest_P, double** P, int* indizes, int ind_length);
int cpget_i(char** dest_P, char** P, int* indizes, int ind_length);
void pset_i(double** dest_P, double** P, int* indizes, int ind_length);

/*------------------------------------------------------------------------------------------------*/

int iremove_entry(int* dest, int* src, int r, int length);
int iremove_value(int* dest, int* src, int v, int length);

/*------------------------------------------------------------------------------------------------*/

double** m_transpose(double** m, long m_size, long m_entry_size);
double** m_create (long m_size, long entry_size);
float** fm_create (long m_size, long entry_size);
int** im_create (long m_size, long entry_size);
char** cm_create (long m_size, long entry_size);
void m_free (double** m);
void fm_free (float** m);
void im_free (int** m);
void cm_free (char** m);
double** m_concatenate (double** m1, int m1_size, double** m2, int m2_size, int entry_size);
void m_to_vec (double* v, double** m, long m_size, long entry_size);
void vec_to_m(double** m, double* v, long m_size, long entry_size);
void im_to_vec (int* v, int** m, int m_size, int entry_size);
void cm_to_vec (char* v, char** m, int m_size, int entry_size);
void m_assign(double** m, long m_size, long entry_size, double val);
void im_assign(int** m, long m_size, long entry_size, int val);
void m_copy (double** dest, double** src, long m_size, long entry_size);
void im_copy (int** dest, int** src, int m_size, int entry_size);
void cm_copy (char** dest, char** src, int m_size, int entry_size);
char** strs_copy(char** strs, int n);

void m_zero (double** m, long m_size, long entry_size);
void im_zero (int** m, long m_size, long entry_size);

void m_print (const char* text, double** m, int m_size, int m_entry_size);

void m_min(int* m_index, int* entry_index, double** m, int m_size, int entry_size);
void m_max(int* m_index, int* entry_index, double** m, int m_size, int entry_size);
/*---------------------------------------------------------------------------------------------*/

int bucket_sort (int* dest, int* x, int x_length, int* bucket, int bucket_size);
/*------------------------------------------------------------------------------------------------*/

double scp (double* fac1, double* fac2, int length);

double diff_scp(double* x1, double* x2, int length);
double l0_diff_scp(double* x1, double* x2, int length);

long* long_array (long length);
int* int_array (long length);
char* char_array (long length);

double* double_array (long length);
float* float_array (long length);
void** pointer_array (long length);
void free_array (void* ptr);
void free_pointers(void** ptrs, int n);
void empty_printf(const char* s, ...);

int is_aligned (void* p);
int align_data (void** pp, int size);

void intersect (int* A, int* B, int A_length, int B_length, int* S, int* S_length, int* I=NULL, int* J=NULL);
void set_diff  (int* A, int* B, int A_length, int B_length, int* X, int* Y, int* X_length, int* Y_length, int* I=NULL, int* J=NULL);

void bin_search (double val, double* x, int* left, int* right, int length);

void int2double  (double* dest, int* src, int length);
void char2double  (double* dest, char* src, int length);
void uchar2double  (double* dest, unsigned char* src, int length);
void double2int  (int* dest, double* src, int length);
void long2double (double* dest, long* src, int length);
void double2long (long* dest, double* src, int length);
void double2char (char* dest, double* src, int length);
void double2uchar (unsigned char* dest, double* src, int length);

void double2float(float* dest, double* src, int length);
void float2double(double* dest, float* src, int length);

void int3tochar(char* dest, char* src, int slength);
void chartoint3(char* dest, char* src, int dlength);

void bin2char(char* dest, char* src, int dlength);
void char2bin(char* dest, char* src, int slength);

void linspace (double* dest, double pmin, double pmax, int count);

int is_little_endian (void);
void toogle_endianess (int* vals, int length);

double equal_error_rate (double* decision, double* cl, int length, double plabel, double nlabel, double* r_b);

#endif
